import sys
import requests
import time
from os import environ
from datetime import datetime
from io import StringIO
from typing import Union, IO, Optional, Tuple
from rdflib import Graph, URIRef, BNode, Literal
from rdflib.namespace import RDF
from codemeta.common import AttribDict, SDO, CODEMETA, license_to_spdx, parse_human_name, generate_uri
from codemeta.parsers.jsonld import parse_jsonld_data

GITAPI_REPO_BLACKLIST=["https://codeberg.org/","http://codeberg.org", "https://git.sr.ht/", "https://bitbucket.com/"]
#it shall be persistent because each new yaml a new invoke of codemetapy is performed and so memory reset
repo_type_cache = {}

def _parse_source(source:str) -> Tuple[str,str,str]:
    source=source.strip("/")
    cleaned_url= source
    scheme=""
    if source.startswith("https://"):
        host = cleaned_url.replace('https://','').split('/')[0]
        scheme="https://"
    else:
        raise ValueError(source + " source url format not recognized!!")
    return cleaned_url, scheme, host
 


def get_repo_kind(source: str) -> Optional[str]:
    source, scheme, host = _parse_source(source)
 
    repo_kind = None
    if "github.com/" in source:
        repo_kind = "github"
    elif "gitlab.com/" in source:
        repo_kind = "gitlab"
    elif f"{scheme}{host}/" not in GITAPI_REPO_BLACKLIST:
        #we have another URL that may or may not be a private gitlab instance, test
        if f"{scheme}{host}/" in repo_type_cache:
            repo_kind = repo_type_cache[f"{scheme}{host}/"]
        else:
            test_url = f"{scheme}{host}/-/manifest.json"  #this seems a relatively cheap way to test if it's a gitlab instance
            response = requests.get(test_url)
            if response.status_code == 200 and response.headers['Content-Type'].startswith("application/json"):
                response = response.json()
                if response['short_name'] == 'GitLab':
                    repo_kind = "gitlab"

    #Populate the cache even when there is a 4xx failure
    repo_type_cache[f"{scheme}{host}/"] = repo_kind

    return repo_kind

def parse(g: Graph, res: Union[URIRef, BNode], source: str, repo_kind:str, args: AttribDict) -> str:
    source, scheme, host = _parse_source(source)
 
    github_suffix=source.replace(scheme + host,'')[1:]
    gitlab_suffix=github_suffix.replace('/', '%2F')
    gitlab_repo_api_url = f"{scheme}{host}/api/v4/projects/{gitlab_suffix}"

    if repo_kind == "github":
        response = rate_limit_get(f"{scheme}api.github.com/repos/{github_suffix}", "github")
        _parse_github(response, g,res,f"{scheme}{host}", args)
    elif repo_kind == "gitlab":
        response = rate_limit_get(gitlab_repo_api_url, "gitlab")
        _parse_gitlab(response, g,res,f"{scheme}{host}", args)  
    else:
        raise ValueError(f"Not a git API, repo_kind={repo_kind}")

    return source

github_crosswalk_table = {
    SDO.codeRepository: "html_url",
    SDO.dateCreated: "created_at",
    SDO.dateModified: "pushed_at",
    SDO.description: "description",
    SDO.name: "name",
}
#"owner": ["owner", "login"],
#"ownerType": ["owner", "type"],  # used to determine if owner is User or Organization

# the same as requests.get(args).json(), but protects against rate limiting
# Adapted from source: https://github.com/KnowledgeCaptureAndDiscovery/somef (MIT licensed)
def rate_limit_get(url:str, repo_kind: Optional[str], backoff_rate=2, initial_backoff=1, **kwargs) -> dict: 
    rate_limited = True
    data = {}
    has_token=False
    if not kwargs: kwargs = {}
    if repo_kind == "github" and 'GITHUB_TOKEN' in environ and environ['GITHUB_TOKEN']:
        if 'headers' not in kwargs: kwargs['headers'] = {}
        kwargs['headers']["Authorization"] = "token " + environ['GITHUB_TOKEN']
        has_token = True
    elif repo_kind == "gitlab" and 'GITLAB_TOKEN' in environ and environ['GITLAB_TOKEN']:
        if 'headers' not in kwargs: kwargs['headers'] = {}
        kwargs['headers']["PRIVATE-TOKEN"] = environ['GITLAB_TOKEN']
        has_token = True
    while rate_limited:
        print(f"Querying {url}")
        response = requests.get(url, **kwargs)
        rate_limit_remaining = int(response.headers.get("RateLimit-Remaining" if repo_kind == "gitlab" else "x-ratelimit-remaining",-1))
        epochtime = int(response.headers.get("RateLimit-Reset" if repo_kind == "gitlab" else  "x-ratelimit-reset",0))
        if rate_limit_remaining > -1 and epochtime > 0:
            date_reset = datetime.fromtimestamp(epochtime)
            print(f"Remaining {repo_kind} API requests: {rate_limit_remaining} ### Next rate limit reset at: {date_reset} (has_token={has_token})")
        else:
            rate_limited = False
        data = response.json()
        if 'message' in data and 'API rate limit exceeded' in data['message']:
            rate_limited = True
            print(f"{repo_kind} API: rate limited. Backing off for {initial_backoff} seconds (has_token={has_token})", file=sys.stderr)
            sys.stderr.flush()
            if initial_backoff > 120:
                raise Exception(f"{repo_kind} API timed out because of rate limiting, giving up... (has_token={has_token})")
            time.sleep(initial_backoff)
            # increase the backoff for next time
            initial_backoff *= backoff_rate
        else:
            response.raise_for_status()
            rate_limited = False
    return data

def _parse_github(response: dict, g: Graph, res: Union[URIRef, BNode], source: str, args: AttribDict):
    """Query and parse from the github API"""
    print(f"    Parsing Github API response",file=sys.stderr)
    users_api_url=f"https://api.github.com/users/"
    
    #repo = response['name']
    for prop, github_key in github_crosswalk_table.items():
        if github_key in response and response[github_key]:
            g.add((res, prop, Literal(response[github_key])))

    if response.get('license') and response['license'].get('spdx_id'):
        g.add((res, SDO.license, Literal(license_to_spdx(response['license']['spdx_id']))))

    if response.get("topics"):
        for topic in response['topics']:
            g.add((res, SDO.keywords, Literal(topic)))

    if response.get("homepage"):
        g.add((res, SDO.url, Literal(response['homepage'])))

    if response.get('has_issues', False) and response.get("html_url"):
        g.add((res, CODEMETA.issueTracker, Literal(response['html_url'] + "/issues")))

    if 'owner' in response:
        owner = response['owner']['login']
        owner_api_url = f"{users_api_url}{owner}"
        response = rate_limit_get(owner_api_url, "github")
        owner_type = response.get("type","").lower()
        owner_res = None
        if owner_type == "user" and response.get('name'):
            firstname, lastname = parse_human_name(response['name'])
            owner_res = URIRef(generate_uri(firstname + "-" + lastname, args.baseuri, prefix="person"))
            g.add((owner_res, RDF.type, SDO.Person))
            g.add((owner_res, SDO.givenName, Literal(firstname)))
            g.add((owner_res, SDO.familyName, Literal(lastname)))
            g.add((res, SDO.author, owner_res))
            g.add((res, SDO.maintainer, owner_res))
            if response.get('company'):
                affil_res = URIRef(generate_uri(response.get('company'), args.baseuri, prefix="org"))
                g.add((affil_res, RDF.type, SDO.Organization))
                g.add((affil_res, SDO.name, Literal(response['company'])))
                g.add((owner_res, SDO.affiliation, affil_res))
        elif owner_type == "organization" and response.get('name'):
            owner_res = URIRef(generate_uri(response.get('name'), args.baseuri, prefix="org"))
            g.add((owner_res, RDF.type, SDO.Organization))
            g.add((owner_res, SDO.name, Literal(response.get('name'))))
            g.add((res, SDO.producer, owner_res))
        if owner_res:
            if response.get('email'):
                g.add((owner_res, SDO.email, Literal(response.get('email'))))
            if response.get('blog'):
                g.add((owner_res, SDO.url, Literal(response.get('blog'))))


gitlab_crosswalk_table = {
    SDO.codeRepository: "web_url",
    SDO.dateCreated: "created_at",
    SDO.dateModified: "last_activity_at",
    SDO.description: "description",
    SDO.name: "name",
    SDO.url: "web_url"
}
def _parse_gitlab(response: dict, g: Graph, res: Union[URIRef, BNode], source, args: AttribDict):
    """Query and parse from the gitlab API"""
    users_api_url = f"{source}/api/v4/users/"
    #Processing start
    for prop, gitlab_key in gitlab_crosswalk_table.items():
        if gitlab_key in response and response[gitlab_key]:
            g.add((res, prop, Literal(response[gitlab_key])))

    if response.get('license') and response['license'].get('nickname'):
        g.add((res, SDO.license, Literal(license_to_spdx(response['license']['nickname']))))
    if response.get("topics"):
        for topic in response['topics']:
            g.add((res, SDO.keywords, Literal(topic)))
    if response.get("homepage"):
        g.add((res, SDO.url, Literal(response['homepage'])))
    elif response.get("web_url"):
        g.add((res, SDO.url, Literal(response['web_url'])))
    if response.get('open_issues_count', False) > 0:
        g.add((res, CODEMETA.issueTracker, Literal(response['_links']['issues'])))

    #https://docs.gitlab.com/ee/api/users.html
    #namespace kind can be just group or user
    owner_id_str=""
    owner_name=""
    user_url=""
    public_mail=""
    if 'namespace' in response and response['namespace']['kind'] == 'user':
        owner_id_str=str(response['namespace']['id'])
        owner_api_url = users_api_url + owner_id_str
        owner_name=response['namespace']['name']
        user_url=response['namespace']['web_url']
    elif 'owner' in response:
        owner_id_str=str(response['owner']['id'])
        owner_api_url = users_api_url + owner_id_str
        response_owner = rate_limit_get(owner_api_url, "gitlab")
        owner_name=response_owner['owner']['name']
        user_url=response_owner['owner']['web_url']
        if response_owner.get('public_email'):
            public_mail = response_owner.get('public_email')
    else:  
        return
    firstname, lastname = parse_human_name(owner_name)
    owner_res = URIRef(user_url)
    g.add((owner_res, RDF.type, SDO.Person))
    g.add((owner_res, SDO.givenName, Literal(firstname)))
    g.add((owner_res, SDO.familyName, Literal(lastname)))
    g.add((owner_res, SDO.url, Literal(user_url)))
    if public_mail != "":
        g.add((owner_res, SDO.email, Literal(public_mail)))
    #Creator considered as author
    response_creator_url_field=user_url
    response_creator_name=owner_name
    if 'creator_id' in response:
     creator_id_str = str(response['creator_id'])
     if creator_id_str != owner_id_str:
        creator_api_url = users_api_url +  creator_id_str
        response_creator = rate_limit_get(creator_api_url, "gitlab")
        response_creator_url_field=response_creator['web_url']
    #Object X must be an rdflib term:  g.add((URIRef(response_creator_url_field), SDO.author, response_creator_name))
    #g.add((res, SDO.maintainer, owner_res))
    #if response_owner.get('work_information'): is like company?



import sys
import os
import datetime
from rdflib import Graph, Namespace, URIRef, BNode, Literal
from rdflib.namespace import RDF, SH
from typing import Union, IO, Sequence, Optional, Tuple
from codemeta.common import init_graph, init_context, CODEMETA, AttribDict,  SDO,  generate_uri

from pyshacl import validate as pyshacl_validate

def validate(g: Graph, res: Union[URIRef,BNode], args: AttribDict, contextgraph: Union[Graph,None] = None) -> Tuple[bool, Graph]:
    """Validates software metadata using SHACL, generates a validation report and adds it to the SoftwareSourceCode metadata via the schema:review property"""
    shacl_file: str = args.validate
    if shacl_file.endswith("ttl"):
        shacl_format="turtle"
    elif shacl_file.endswith(("json","jsonld")):
        shacl_format="json-ld"
    else:
        raise ValueError(f"Expect ttl or json file for SHACL ({args.validate}), unable to determine from extension")
    shacl_graph = Graph()
    shacl_graph.parse(args.validate, format=shacl_format)
    conforms, results_graph, _ = pyshacl_validate(data_graph=g, shacl_graph=shacl_graph, ont_graph=contextgraph, abort_on_first=False, allow_infos=True, allow_warnings=True)
    counter = 0
    review = URIRef(generate_uri(None, args.baseuri,prefix="validation"))
    g.add((review, RDF.type, SDO.Review))
    g.add((review, SDO.author, Literal(f"codemetapy validator using {os.path.basename(shacl_file)}")))
    g.add((review, SDO.datePublished, Literal(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"))))
    name = g.value(res,SDO.name)
    if not name: name = "unnamed software"
    version = g.value(res,SDO.version)
    if not version: version = "(unknown version)"
    g.add((review, SDO.name, Literal(f"Automatic software metadata validation report for {name} {version}")))
    messages = []
    warnings = 0
    violations = 0
    info = 0
    for node ,_,_ in results_graph.triples((None,SH.focusNode,res)):
        if (node, RDF.type, SH.ValidationResult) in results_graph:
            severity = ""
            if (node, SH.resultSeverity, SH.Violation) in results_graph:
                severity = "Violation"
                violations += 1
            elif (node, SH.resultSeverity, SH.Warning) in results_graph:
                severity = "Warning"
                warnings += 1
            elif (node, SH.resultSeverity, SH.Info) in results_graph:
                severity = "Info"
                info += 1
            else:
                severity = "Unknown"
            cause = ""
            if (node, SH.sourceConstraintComponent, SH.ClassConstraintComponent) in results_graph:
                cause = "The metadata does express this currently, but the wrong class is used."
            elif (node, SH.sourceConstraintComponent, SH.DatatypeConstraintComponent) in results_graph:
                cause = "The metadata does express this currently, but the wrong datatype is used"
            elif (node, SH.sourceConstraintComponent, SH.NodeKindConstraintComponent) in results_graph:
                cause = "The metadata does express this currently, but the wrong node kind is used (probably URI vs literal)"
            elif (node, SH.sourceConstraintComponent, SH.OrConstraintComponent) in results_graph:
                cause = "The metadata does express this currently, but something is wrong in the way it is expressed. Is the type/class valid?"
            elif (node, SH.sourceConstraintComponent, SH.QualifiedMinCountConstraintComponent) in results_graph:
                cause = "This is missing in the metadata"
            elif (node, SH.sourceConstraintComponent, SH.QualifiedMaxCountConstraintComponent) in results_graph:
                cause = "The metadata expresses this multiple times and is too ambiguous"
            elif (node, SH.sourceConstraintComponent, SH.MinCountConstraintComponent) in results_graph:
                cause = "This is missing in the metadata"
            elif (node, SH.sourceConstraintComponent, SH.MaxCountConstraintComponent) in results_graph:
                cause = "The metadata expresses this multiple times and is too ambiguous"
            if cause: cause = f"({cause})"

            #path = results_graph.value(node, SH.resultPath)
            msg = results_graph.value(node, SH.resultMessage)
            if msg:
                counter +=1 
                print(f"VALIDATION {str(res)} #{counter}: {severity}: {str(msg)} {cause}", file=sys.stderr)
                messages.append(f"{counter}. {severity}: {msg} {cause}")
    head = args.textv  + "\n\n" if args.textv else ""
    if messages:
        if conforms:
            if warnings:
                head += f"Validation of {name} {version} was successful (score=3/5), but there are some warnings which should be addressed:"
                g.add((review, SDO.reviewRating, Literal(3)))
            else:
                head += f"Validation of {name} {version} was successful (score=4/5), but there are some remarks which you may or may not want to address:"
                g.add((review, SDO.reviewRating, Literal(4)))
        else:
            if violations > 3:
                g.add((review, SDO.reviewRating, Literal(0)))
                score = "(score 0/5)"
            elif violations > 1 or warnings > 5:
                g.add((review, SDO.reviewRating, Literal(1)))
                score = "(score 1/5)"
            else:
                g.add((review, SDO.reviewRating, Literal(2)))
                score = "(score 2/5)"
            head += f"Validation of {name} {version} failed {score} due to one or more requirement violations:"
        g.add((review, SDO.reviewBody, Literal(head + "\n\n" + "\n".join(messages))))
    else:
        g.add((review, SDO.reviewBody, Literal("Validates perfectly, no further remarks!")))
        g.add((review, SDO.reviewRating, Literal(5)))
    g.add((res, SDO.review, review))
    return conforms, results_graph

def get_validation_report(g: Graph, res: Union[Sequence,URIRef,BNode]) -> Optional[str]:
    """Get the text of an existing validation report for the given resource"""
    if (res,RDF.type,SDO.Review):
        return g.value(res,SDO.reviewBody)
    else:
        for _,_,review in g.triples((res,SDO.review,None)):
            return g.value(review,SDO.reviewBody)

            

    



import csv
import os.path
from collections import defaultdict

class CWKey:
    """Crosswalk Keys, corresponds with header label in crosswalk.csv, as
    provided by the CodeMeta project"""
    PROP = "Property"
    PARENT = "Parent Type"
    TYPE = "Type"
    DESCRIPTION = "Description"
    PYPI = "Python Distutils (PyPI)"
    DEBIAN = "Debian Package"
    R = "R Package Description"
    NODEJS = "NodeJS"
    MAVEN = "Java (Maven)"
    DOAP = "DOAP"

def readcrosswalk(sourcekeys=(CWKey.PYPI,CWKey.DEBIAN,CWKey.NODEJS, CWKey.MAVEN)):
    """Read the crosswalk.csv as provided by codemeta into memory"""
    #pylint: disable=W0621
    crosswalk = defaultdict(dict)
    #pip may output things differently than recorded in distutils/setup.py, so we register some aliases:
    crosswalk[CWKey.PYPI]["home-page"] = "url"
    crosswalk[CWKey.PYPI]["summary"] = "description"
    props = {}
    crosswalkfile = os.path.join(os.path.dirname(__file__), 'schema','crosswalk.csv')
    with open(crosswalkfile, 'r', encoding="utf-8") as f:
        reader = csv.DictReader(f)
        for row in reader:
            props[row[CWKey.PROP]] = {"PARENT": row[CWKey.PARENT], "TYPE": row[CWKey.TYPE], "DESCRIPTION": row[CWKey.DESCRIPTION] }
            for sourcekey in sourcekeys:
                if row[sourcekey]:
                    for key in [ x.strip().lower() for x in row[sourcekey].split("/") ]:
                        crosswalk[sourcekey][key] = row[CWKey.PROP]

    return props, crosswalk



import sys
import json
import os.path
from typing import Union, IO
from rdflib import Graph, URIRef, BNode, Literal
from rdflib.namespace import RDF
from codemeta.common import AttribDict, add_triple, CODEMETA, SOFTWARETYPES, add_authors, SDO, COMMON_SOURCEREPOS, SOFTWARETYPES, generate_uri
from codemeta.crosswalk import readcrosswalk, CWKey


def parse_sourcerepo(value):
    """npm allows some shortcuts, resolve them"""
    if value.startswith("github:"):
        value = "https://github.com/" + value[len("github:"):]
    elif value.startswith("gitlab:"):
        value = "https://gitlab.com/" + value[len("gitlab:"):]
    elif value.startswith("bitbucket:"):
        value = "https://bitbucket.com/" + value[len("bitbucket:"):]
    value = value.replace("git://","https://").replace("git+ssh://","https://") #always prefer https in URLs
    return value

def parse_nodejs(g: Graph, res: Union[URIRef, BNode], file: IO , crosswalk, args: AttribDict):
    data = json.load(file)
    iswebapp = False
    foundrepo = False
    for key, value in data.items():
        if key.lower() in crosswalk[CWKey.NODEJS]:
            if key == 'name' and value:
                #remove the 'scope' from the name
                if value[0] == '@' and value.find('/') != -1:
                    value = value.split('/')[1]
                add_triple(g, res, key, value, args)
            elif key == 'bugs':
                if isinstance(value,dict) and 'url' in value:
                    g.add((res, CODEMETA.issueTracker, Literal(value['url'])))
                    if 'email' in value:
                        g.add((res, SDO.email, Literal(value['email'])))
                elif isinstance(value, str):
                    g.add((res, CODEMETA.issueTracker, Literal(value)))
            elif key == 'license':
                if isinstance(value, dict) and 'type' in value:
                    add_triple(g, res, "license", value['type'], args)
                elif isinstance(value, (list,tuple)):
                    for item in value:
                        if isinstance(item, dict) and 'type' in item:
                            add_triple(g, res, "license", item['type'], args)
                        elif isinstance(item, str):
                            add_triple(g, res, "license", item, args)
                elif isinstance(value, str):
                    add_triple(g, res, "license", value, args)
            elif key == 'keywords':
                if isinstance(value, (list,tuple)):
                    for keyword in value:
                        add_triple(g, res, "keywords", keyword, args)
                else:
                    print("WARNING: keywords in package.json should be a list",file=sys.stderr)
            elif key == 'private' and value:
                    print("WARNING: private=true was set on package.json! This package is marked not to be published!",file=sys.stderr)
            elif key == 'repository':
                if isinstance(value, str):
                    value = parse_sourcerepo(value)
                    add_triple(g, res, "codeRepository", value, args)
                    foundrepo = True
                elif isinstance(value, dict) and 'url' in value:
                    value = parse_sourcerepo(value['url'])
                    add_triple(g, res, "codeRepository", value, args)
                    foundrepo = True
            elif key == 'homepage':
                for sourcerepo in COMMON_SOURCEREPOS:
                    if value.startswith(sourcerepo) and not foundrepo:
                        #catch if we're describing the source code repo instead
                        add_triple(g, res, "codeRepository", value, args)
                        break
                add_triple(g, res, "url", value, args)
            elif key == 'author':
                #npm prescribes that author is only one person
                if isinstance(value, dict) and 'name' in value:
                    authors = add_authors(g, res, value['name'], single_author=True, mail=value.get("email"), baseuri=args.baseuri)
                    if authors and 'url' in value:
                        g.add((authors[0], SDO.url, Literal(value['url'])))
                elif isinstance(value, str):
                    #npm allows strings like "Barney Rubble <b@rubble.com> (http://barnyrubble.tumblr.com/)"
                    #our add_authors function can handle that directly
                    add_authors(g, res, value, single_author=True, baseuri=args.baseuri)
            elif key == 'contributors':
                if isinstance(value, dict) and 'name' in value:
                    authors = add_authors(g, res, value['name'], property=SDO.contributor, single_author=True, mail=value.get("email"), baseuri=args.baseuri)
                elif isinstance(value, str):
                    add_authors(g, res, value, property=SDO.contributor, baseuri=args.baseuri)
                elif isinstance(value, (list,tuple)):
                    for value in value:
                        authors = add_authors(g, res, value['name'], property=SDO.contributor, single_author=True, mail=value.get("email"), baseuri=args.baseuri)
            elif key in ('dependencies','devDependencies'):
                if isinstance(value, dict):
                    for key, versioninfo in value.items():
                        depres = URIRef(generate_uri(key+versioninfo, baseuri=args.baseuri,prefix="dependency"))
                        g.add((depres, RDF.type, SDO.SoftwareApplication))
                        g.add((depres, SDO.name, Literal(key)))
                        g.add((depres, SDO.identifier, Literal(key)))
                        g.add((depres, SDO.version, Literal(versioninfo)))
                        g.add((res,  CODEMETA.softwareRequirements, depres))
                    #detect some common web application frameworks or other
                    #dependencies that indicate this is a web-app
                    iswebapp = iswebapp or 'react' in value or 'vue' in value or 'sitemap' in value or  'gatsby' in value
            elif key in ('bundledDependencies','peerDependencies'):
                pass #ignore
            elif key == 'bin':
                #note: assuming CommandLineApplication may be a bit presumptuous here
                if isinstance(value, dict) and 'name' in value and args.with_stypes:
                    for progname, execname in value.items():
                        sapp = URIRef(generate_uri(key, baseuri=args.baseuri,prefix="commandlineapplication"))
                        g.add((sapp, RDF.type, SOFTWARETYPES.CommandLineApplication))
                        g.add((sapp, SDO.name, progname))
                        g.add((sapp, SOFTWARETYPES.executableName, os.path.basename(execname)))
                        g.add((res, SDO.targetProduct, sapp))
                elif isinstance(value, str) and args.with_stypes:
                    sapp = URIRef(generate_uri(data['name'], baseuri=args.baseuri,prefix="commandlineapplication"))
                    g.add((sapp, RDF.type, SOFTWARETYPES.CommandLineApplication))
                    g.add((sapp, SDO.name, data['name'])) #from parent
                    g.add((sapp, SOFTWARETYPES.executableName, os.path.basename(value)))
                    g.add((res, SDO.targetProduct, sapp))
            elif key == 'engines':
                if isinstance(value, dict):
                    for key, versioninfo in value.items():
                        g.add((res, SDO.runtimePlatform, Literal(key + " " + versioninfo)))
            else:
                key = crosswalk[CWKey.NODEJS][key.lower()]
                add_triple(g, res, key, value, args)

    if 'devDependencies' in data and "typescript" in data['devDependencies']:
        g.add((res, SDO.programmingLanguage, Literal("Typescript")))
    else:
        g.add((res, SDO.programmingLanguage, Literal("Javascript")))

    if args.with_stypes and 'browser' in data or 'browserslist' in data or iswebapp:
        #assume this is a web-application
        sapp = URIRef(generate_uri(data['name'], baseuri=args.baseuri,prefix="webapplication"))
        g.add((sapp, RDF.type, SDO.WebApplication))
        g.add((sapp, SDO.name, Literal(data['name']))) #from parent
        g.add((sapp, SDO.version, Literal(data['version']))) #from parent
        g.add((res, SDO.targetProduct, sapp))



import sys
from typing import Union
from rdflib import Graph, URIRef, BNode, Literal
from rdflib.namespace import RDF
from codemeta.common import AttribDict, add_triple, CODEMETA, SOFTWARETYPES, SDO, SOFTWARETYPES, generate_uri
from codemeta.crosswalk import readcrosswalk, CWKey

def parse_debian(g: Graph, res: Union[URIRef, BNode], lines, crosswalk, args: AttribDict):
    """Parses apt show output and converts to codemeta"""
    if crosswalk is None:
        _, crosswalk = readcrosswalk((CWKey.DEBIAN,))

    name = None
    interfacetype = None
    description = ""
    parsedescription = False
    for line in lines:
        if parsedescription and line and line[0] == ' ':
            description += line[1:] + " "
        else:
            try:
                key, value = (x.strip() for x in line.split(':',1))
            except:
                continue
            if key == "Origin":
                if value == "Debian":
                    provider = URIRef("https://www.debian.org")
                    g.add((provider, RDF.type, SDO.Organization))
                    g.add((provider, SDO.name, Literal("The Debian Project")))
                    g.add((provider, SDO.url, Literal("https://www.debian.org")))
                    g.add((res,SDO.provider,provider))
                elif value == "Ubuntu":
                    provider = URIRef("https://ubuntu.com")
                    g.add((provider, RDF.type, SDO.Organization))
                    g.add((provider, SDO.name, Literal("Ubuntu")))
                    g.add((provider, SDO.url, Literal("https://ubuntu.com")))
                    g.add((res,SDO.provider,provider))
                else:
                    print(f"WARNING: Don't know how to convert Origin: {value}",file=sys.stderr)
            elif key == "Depends":
                for dependency in value.split(","):
                    dependency = dependency.strip().split(" ")[0].strip()
                    if dependency:
                        depnode = URIRef(generate_uri(dependency, baseuri=args.baseuri,prefix="dependency"))
                        g.add((depnode, RDF.type, SDO.SoftwareApplication))
                        g.add((depnode, SDO.identifier, Literal(dependency)))
                        g.add((depnode, SDO.name, Literal(dependency)))
                        g.add((res, CODEMETA.softwareRequirements, depnode))
            elif key == "Section":
                #attempt to make an educated guess for the audience and interface type
                if "libs" in value or "libraries" in value:
                    if args.with_stypes:
                        interfacetype = SOFTWARETYPES.SoftwareLibrary
                        add_triple(g, res, "audience", "Developers", args)
                elif "utils" in value or "text" in value:
                    if args.with_stypes:
                        interfacetype = SOFTWARETYPES.CommandLineApplication
                elif "devel" in value:
                    add_triple(g, res, "audience", "Developers", args)
                elif "science" in value:
                    add_triple(g, res, "audience", "Researchers", args)
            elif key == "Description":
                parsedescription = True
                description = value + "\n\n"
            elif key == "Homepage":
                g.add((res, SDO.url, Literal(value)))
            elif key == "Version":
                g.add((res, SDO.version, Literal(value)))
            elif key.lower() in crosswalk[CWKey.DEBIAN]:
                if key == "Package":
                    name = value
                else:
                    key = crosswalk[CWKey.DEBIAN][key.lower()]
                    add_triple(g, res, key, value, args)
            else:
                print("WARNING: No translation for APT key " + key,file=sys.stderr)
    if name:
        g.add((res, SDO.name, Literal(name)))
        g.add((res, SDO.identifier, Literal(name)))
    else:
        print("No name found for package, should not happen",file=sys.stderr)
        return False
    if description:
        g.add((res, SDO.description, Literal(description)))
    if interfacetype and args.with_stypes:
        sapp = URIRef(generate_uri(name, baseuri=args.baseuri,prefix=str(interfacetype).lower()))
        g.add((sapp, RDF.type, interfacetype))
        g.add((sapp, SDO.name, name))
        g.add((res, SDO.targetProduct, sapp))




import sys
import json
import os.path
from typing import Union, Iterator
import requests
import yaml
from rdflib import Graph, URIRef, BNode, Literal
from rdflib.namespace import RDF
from codemeta.common import AttribDict, SDO, generate_uri, add_authors, get_last_component
from codemeta.parsers.jsonld import parse_jsonld_data
from bs4 import BeautifulSoup


class MiddlewareObstructionException(Exception):
    pass

def detect_type(data):
    if "@context" in data and "@type" in data:
        value = data['@context']
        if isinstance(value,str):
            if value.find("schema.org") != -1 or value.find("codemeta") != -1:
                return  "schema"
        elif isinstance(value,list):
            if any( x.find("schema.org") != -1 or x.find("codemeta") != -1 for x in value ):
                return "schema"
    elif isinstance(data, dict) and 'openapi' in data:
        return "openapi"
    return None

def get_meta(soup, *keys, default=""):
    for key in keys:
        for value in  soup.find("head").find_all("meta", itemprop=key):
            if value.get('content'):
                return value.get('content')
        for value in  soup.find("head").find_all("meta", property=key):
            if value.get('content'):
                return value.get('content')
        for value in soup.find("head").find_all("meta", {"name": key }):
            if value.get('content'):
                return value.get('content')
    return default


def get_soup(soup,elementname, attribname=None):
    for e in soup.find_all(elementname):
        if attribname:
            v = e.get(attribname)
            return v
        elif e.text:
            return e.text
    return ""

def parse_clam(soup):
    return {
        "name": get_soup(soup,"clam","name"),
        "url": get_soup(soup, "clam", "baseurl"),
        "description": get_soup(soup, "description"),
        "author": get_soup(soup, "author"),
        "provider": get_soup(soup, "affiliation"),
        "email": get_soup(soup, "email"),
        "version": get_soup(soup, "version"),
        "license": get_soup(soup, "license")
    }

def detect_sso_middleware(r: requests.Response):
    return r.history and 'location' in r.history[-1].headers and r.history[-1].headers['location'].lower().find("shibboleth") != -1

def add_missing_url_scheme(data, original_url: str):
    """Add URL scheme when missing"""
    if isinstance(data, dict):
        return {k: add_missing_url_scheme(v, original_url) for k, v in data.items()}
    elif isinstance(data, (list,tuple)):
        return [ add_missing_url_scheme(v, original_url) for v in data ]
    elif isinstance(data, str) and data.startswith("//"):
        return original_url.split("/")[0] + data
    return data

def parse_web(g: Graph, res: Union[URIRef, BNode], url, args: AttribDict) -> Iterator[Union[URIRef,BNode,None]]:
    r = requests.get(url, headers={ "Accept": "application/json+ld;q=1.0,application/json;q=0.9,application/x-yaml;q=0.8,application/xml;q=0.7;text/html;q=0.6;text/plain;q=0.1" })
    r.raise_for_status()
    contenttype = r.headers.get('content-type',"").split(';')[0].strip()
    print(f"    Service replied with content-type {contenttype}",file=sys.stderr)
    datatype = None
    data = None


    if contenttype in ("application/json", "application/ld+json") or url.endswith(".json") or url.endswith(".jsonld"):
        print("    Parsing json...",file=sys.stderr)
        data = json.loads(r.text)
    elif contenttype in ("application/x-yaml", "text/yaml") or url.endswith(".yml") or url.endswith(".yaml"):
        print("    Parsing yaml...",file=sys.stderr)
        #may be OpenAPI
        with open(r.text,'r', encoding="utf-8") as f:
            data = yaml.load(f, yaml.Loader)
    elif contenttype == "text/html":
        if detect_sso_middleware(r):
            #we've been redirected to a login page directly
            #so we can't extract any useful metadata
            #we add a dummy entry:
            raise MiddlewareObstructionException(f"Unable to extract metadata from {url} because it immediately redirects to an external (SSO) login page rather than a proper landing page")
        #normal behaviour
        print("    Parsing html...",file=sys.stderr)
        soup = BeautifulSoup(r.text, 'html.parser')
        scriptblock = soup.find("script", {"type":"application/ld+json"})
        if scriptblock:
            #Does the site provide proper JSON-LD metadata itself?
            print("    Found a json-ld script block",file=sys.stderr)
            data = json.loads("".join(scriptblock.contents))            
        else:
            print("    Parsing site metadata",file=sys.stderr)
            name = get_meta(soup, "schema:name", "og:site_name", "og:title", "twitter:title")
            if not name and soup.title:
                name = soup.title.text
                name = name.strip()

            if args.with_stypes:
                targetres = URIRef(generate_uri(name, baseuri=args.baseuri,prefix="webapplication"))
            else:
                targetres = res
            if name:
                g.add((targetres, SDO.name, Literal(name)))

            targetrestype = SDO.WebApplication
            for e in (soup.find("head"), soup.find("html")):
                itemtype = e.get("itemtype")
                if itemtype in ("https://schema.org/WebApplication", "http://schema.org/WebApplication"):
                    targetrestype = SDO.WebApplication
                elif itemtype in ("https://schema.org/WebPage", "http://schema.org/WebPage"):
                    targetrestype = SDO.WebPage
                elif itemtype in ("https://schema.org/WebSite", "http://schema.org/WebSite"):
                    targetrestype = SDO.WebSite
                elif itemtype in ("https://schema.org/WebAPI", "http://schema.org/WebAPI"):
                    targetrestype = SDO.WebAPI
                if itemtype:
                    break

            g.add((targetres, RDF.type, targetrestype))
            g.add((targetres, SDO.url, Literal(get_meta(soup, "og:url", "url", default=url))))

            v = get_meta(soup, "schema:description", "og:description", "twitter:description", "description")
            if v: g.add((targetres, SDO.description, Literal(v)))

            v = get_meta(soup, "schema:url")
            if v: g.add((targetres, SDO.url, Literal(add_missing_url_scheme(v,url))))

            v = get_meta(soup, "schema:thumbnailUrl", "og:image", "twitter:image", "thumbnail")
            if v: g.add((targetres, SDO.thumbnailUrl, Literal(add_missing_url_scheme(v,url))))

            v = get_meta(soup, "schema:author", "author")
            if v: add_authors(g, targetres, v, baseuri=args.baseuri)

            v = get_meta(soup, "schema:keywords", "keywords")
            if v:
                for item in v.split(","):
                    item = item.strip()
                    if item: g.add((targetres, SDO.keywords, Literal(item)))

            yield targetres
            data = None
    elif contenttype in ("application/xml", "text/xml"):
        soup = BeautifulSoup(r.text, 'xml')
        if soup.find("clam") and args.with_stypes:
            print("    Parsing CLAM metadata",file=sys.stderr)
            clamdata = parse_clam(soup)
            for restype in (SDO.WebApplication, SDO.WebAPI):
                assert clamdata['name']
                targetres = URIRef(generate_uri(clamdata['name'], baseuri=args.baseuri,prefix=get_last_component(str(restype).lower())))
                g.add((targetres, RDF.type, restype))
                if clamdata['name']:
                    g.add((targetres, SDO.name, Literal(clamdata['name'])))
                if clamdata['description']:
                    g.add((targetres, SDO.description, Literal(clamdata['description'])))
                if clamdata['author']:
                    add_authors(g, targetres, clamdata['author'], baseuri=args.baseuri)
                if clamdata['email']:
                    g.add((targetres, SDO.email, Literal(clamdata['email'])))
                if clamdata['url']:
                    g.add((targetres, SDO.url, Literal(clamdata['url'])))
                    if restype == SDO.WebAPI:
                        g.add((targetres, SDO.documentation, Literal(os.path.join(clamdata['url'],"info")))) #Proposed in schemaorg/schemaorg#1423
                else:
                    g.add((targetres, SDO.url, Literal(url)))
                if clamdata['provider']:
                    g.add((targetres, SDO.provider, Literal(clamdata['provider'])))
                if clamdata['version']:
                    g.add((targetres, SDO.version, Literal(clamdata['version'])))
                yield targetres
            data = None
        else:
            print("    Remote returned unrecognized XML",file=sys.stderr)
    else:
        print(f"    Remote returned unknown contenttype: {contenttype}",file=sys.stderr)

    if data:
        datatype = detect_type(data)
        if datatype == 'schema':
            if args.with_stypes:
                targetres = URIRef(generate_uri(baseuri=args.baseuri,prefix="webapplication"))
            else:
                targetres = res
            data = add_missing_url_scheme(data, url)
            parse_jsonld_data(g, targetres, data, args)
            yield targetres
        elif datatype == 'openapi':
            raise NotImplementedError #TODO
        else:
            print(f"    Unable to detect data type of data returned by {url}",file=sys.stderr)



import sys
import os
import importlib
import re
from typing import Union, IO
if sys.version_info.minor < 8:
    import importlib_metadata #backported
else:
    import importlib.metadata as importlib_metadata #python 3.8 and above: in standard library

from rdflib import Graph, URIRef, BNode, Literal
from rdflib.namespace import RDF
from codemeta.common import AttribDict, add_triple, CODEMETA, SOFTWARETYPES, add_authors, SDO, COMMON_SOURCEREPOS, generate_uri, get_last_component
from codemeta.crosswalk import readcrosswalk, CWKey
import pyproject_parser
import pep517.meta



def splitdependencies(s: str):
    """Split a string of multiple (python) dependencies"""
    begin = 0
    depth = 0
    for i, c in enumerate(s):
        if c == '(':
            depth += 1
        elif c == ')':
            depth -= 1
        if c == ',' and depth == 0:
            yield s[begin:i].strip()
            begin = i + 1
    if s[begin:].strip():
        yield s[begin:].strip()


def parsedependency(s: str):
    """Parses a pip dependency specification, returning the identifier, version"""
    end = min(
        s.find(" ") if s.find(" ") != -1 else 999999,
        s.find(";") if s.find(";") != -1 else 999999,
        s.find(">") if s.find(">") != -1 else 999999,
        s.find("!") if s.find("!") != -1 else 999999,
        s.find("~") if s.find("~") != -1 else 999999,
        s.find("=") if s.find("=") != -1 else 999999,
        s.find("^") if s.find("^") != -1 else 999999
    )
    if end != 999999:
        identifier = s[:end]
    else:
        return s, ""

    versionbegin = -1
    for i, c in enumerate(s[end:]):
        if c not in ('>','=','~','!',' '):
            versionbegin = end + i
            break
    if versionbegin != -1:
        operator = s[end:versionbegin].strip()
        if operator in ("=","=="):
            version = s[versionbegin:].strip()
        else:
            version = operator + " " + s[versionbegin:].strip()
    else:
        version = ""
    if ';' in version:
        #there may be extra qualifiers after the version which we strip (like ; sys_platform=)
        version = version.split(";")[0]
    return identifier, version.strip("[]() -.,:")


def parse_classifier(value: str, g: Graph, res: Union[URIRef, BNode], crosswalk, args: AttribDict):
    fields = [ x.strip() for x in value.strip().split('::') ]
    classifier = fields[0]
    pipkey = f"classifiers['{classifier}']".lower()
    if pipkey in crosswalk[CWKey.PYPI]:
        key = crosswalk[CWKey.PYPI][pipkey]
        det = " > " if key != "programmingLanguage" else " "
        value = det.join(fields[1:])
        add_triple(g, res, "runtimePlatform" if key == "programmingLanguage" else key, value, args)
    elif classifier.lower() in crosswalk[CWKey.PYPI]:
        key = crosswalk[CWKey.PYPI][classifier.lower()]
        det = " > " if key != "programmingLanguage" else " "
        value = det.join(fields[1:])
        add_triple(g, res, "runtimePlatform" if key == "programmingLanguage" else key, value, args)
    elif classifier == "Intended Audience":
        add_triple(g, res, "audience", " > ".join(fields[1:]), args)
    else:
        print("NOTICE: Classifier "  + fields[0] + " has no translation",file=sys.stderr)

def parse_url(label: str, url: str, g: Graph, res: Union[URIRef, BNode], crosswalk, args: AttribDict):
    #label is free so we do some educated guessed
    #(see https://packaging.python.org/en/latest/specifications/core-metadata/#project-url-multiple-use)
    label = label.lower()
    if "repository" in label or label in ('git','github','code','sourcecode',"source"):
        add_triple(g, res, "codeRepository", url, args)
    elif label in ("bug tracker","issue tracker","tracker", "issues"):
        add_triple(g, res, "issueTracker", url, args)
    elif label in ("documentation", "docs","api reference","reference"):
        add_triple(g, res, "softwareHelp", url, args)
    elif label == "readme":
        add_triple(g, res, "readme", url, args)
    elif label in ("build","build instructions","installation"):
        add_triple(g, res, "buildInstructions", url, args)
    elif label in ("release notes","releases"):
        add_triple(g, res, "releaseNotes", url, args)
    elif label in ("continous integration", "ci","tests"):
        add_triple(g, res, "contIntegration", url, args)
    else:
        add_triple(g, res, "url", url, args)

def metadata_from_pyproject(pyproject: pyproject_parser.PyProject):
    if pyproject.project and 'name' in pyproject.project:
        return pyproject.project
    if pyproject.tool and 'name' in list(pyproject.tool.values())[0]:
        return list(pyproject.tool.values())[0]
    return None

#pylint: disable=W0621
def parse_python(g: Graph, res: Union[URIRef, BNode], packagename: str, crosswalk, args: AttribDict):
    """Parses python package metadata and converts it to codemeta

    Parameters:
    * `packagename` - Either 1) a name of a package that is installed or available in the current working directory, or 2) path to a pyproject.toml file 
    """
    if crosswalk is None:
        _, crosswalk = readcrosswalk((CWKey.PYPI,))
    if args.exactplatformversion:
        g.add((res, SDO.runtimePlatform, Literal("Python " + str(sys.version_info.major) + "." + str(sys.version_info.minor) + "." + str(sys.version_info.micro))))
    else:
        g.add((res, SDO.runtimePlatform, Literal("Python 3")))

    prevdir = None
    pkg = None
    if os.path.basename(packagename) == "pyproject.toml":
        print(f"Loading metadata from {packagename} via pyproject-parser",file=sys.stderr) #pylint: disable=W0212
        metadata = None
        try:
            pyproject = pyproject_parser.PyProject.load(packagename)
            metadata = metadata_from_pyproject(pyproject)
            if not metadata:
                print(f"Failed to find complete enough metadata in pyproject.toml",file=sys.stderr)
        except Exception as e:
            print(f"Failed to process {packagename} via pyproject-parser: {str(e)}",file=sys.stderr)
       
        if not metadata:
            print(f"Fallback: Loading metadata from {packagename} via PEP517",file=sys.stderr) #pylint: disable=W0212
            try:
                pkg = pep517.meta.load(os.path.dirname(packagename))
            except Exception as e:
                print(f"Failed to process {packagename} via PEP517: {str(e)}",file=sys.stderr)
                sys.exit(4)
            packagename = pkg.name
            metadata = pkg.metadata.items()
        else:
            metadata = metadata.items()
    else:
        print(f"Loading metadata from {packagename} via importlib.metadata",file=sys.stderr) #pylint: disable=W0212
        try:
            pkg = importlib_metadata.distribution(packagename)
        except importlib_metadata.PackageNotFoundError:
            if '/' in packagename:
                dirname = os.path.dirname(packagename)
                prevdir = os.getcwd()
                os.chdir(dirname)
                packagename = os.path.basename(packagename)
            #fallback if package is not installed but in local directory:
            context = importlib_metadata.DistributionFinder.Context(name=packagename,path='.')
            try:
                pkg = next(importlib_metadata.MetadataPathFinder().find_distributions(context))
            except StopIteration:
                print(f"No such python package: {packagename}",file=sys.stderr)
                if prevdir: os.chdir(prevdir)
                sys.exit(4)
        if isinstance(pkg._path, str):
            print(f"Found metadata in {pkg._path}",file=sys.stderr) #pylint: disable=W0212
        metadata = pkg.metadata.items()
            #if prevdir: os.chdir(prevdir)

    for key, value in metadata:
        if key == "Classifier": #importlib.metadata
            parse_classifier(value, g, res, crosswalk, args)
        elif key == "classifiers" and isinstance(value, (list,tuple)): #pyproject
            for e in value:
                parse_classifier(e, g, res, crosswalk, args)
        else:
            if key == "authors" and isinstance(value, (list,tuple)): #pyproject
                for e in value:
                    if isinstance(e, str):
                        add_authors(g, res, e, single_author=True, baseuri=args.baseuri)
                    elif isinstance(e, dict) and 'name' in e:
                        add_authors(g, res, e['name'], single_author=True, mail=e.get('mail',""), baseuri=args.baseuri)
            elif key == "Author" and pkg: #distutils
                add_authors(g, res, value, single_author=args.single_author, mail=pkg.metadata.get("Author-email",""), baseuri=args.baseuri)
            elif key == "Author-email":
                continue #already handled by the above
            elif key == "Project-URL":
                if ',' in value:
                    label, url = value.split(",",1) #according to spec
                    label = label.strip()
                    url = url.strip()
                    parse_url(label, url, g, res, crosswalk, args)
                else:
                    #input probably not according to spec
                    add_triple(g, res, "url", value, args) 
            elif key == "urls" and isinstance(value,dict): #pyproject.toml
                for label, url in value.items():
                    parse_url(label, url, g, res, crosswalk, args)
            elif key == 'repository':
                add_triple(g, res, "codeRepository", value, args)
            elif key == 'dependencies':
                if isinstance(value, dict):
                    for k,v in value.items():
                        add_dependency(g, res, f"{k} {v}", args)
                elif isinstance(value, list):
                    for e in value:
                        #                      v-- we may get a pyproject_parser.Requirement() instance here, so convert to str()
                        add_dependency(g, res, str(e), args)
            elif key == 'description' and not pkg: #pyproject.toml
                #note: in distutils this is called 'summary' and we don't want the Description field there because it contains the whole README 
                add_triple(g, res, 'description', value, args)
            elif key in ('gui_scripts','gui-scripts') and isinstance(value, dict): #pyproject.toml, flit, poetry?
                for script_name, rawentrypoint in value.items():
                    module_name = rawentrypoint.strip().split(':')[0]
                    interfacetype = SOFTWARETYPES.DesktopApplication
                    add_entrypoint(g, res, script_name, module_name, interfacetype, args)
            elif key == 'scripts' and isinstance(value, dict): #pyproject.toml, poetry
                for script_name, rawentrypoint in value.items():
                    module_name = rawentrypoint.strip().split(':')[0]
                    #we don't really know for sure if it's a CLI or GUI application, assume CLI
                    interfacetype = SOFTWARETYPES.CommandLineApplication
                    add_entrypoint(g, res, script_name, module_name, interfacetype, args)
            elif key.lower() in crosswalk[CWKey.PYPI]:
                add_triple(g, res, crosswalk[CWKey.PYPI][key.lower()], value, args)
                if crosswalk[CWKey.PYPI][key.lower()] == "url":
                    for v in COMMON_SOURCEREPOS:
                        if value.startswith(v):
                            add_triple(g, res, "codeRepository", value, args)
                            if not args.baseuri:
                                prefuri = value
                            break
            else:
                print("WARNING: No translation for distutils or pyproject.toml key " + key,file=sys.stderr)

    if pkg and pkg.requires:
        for value in pkg.requires:
            add_dependency(g, res, value, args)

    test_and_set_identifier(g, res, args)

    if args.with_stypes and pkg:
        found = False
        for rawentrypoint in pkg.entry_points:
            if rawentrypoint.group == "console_scripts":
                interfacetype = SOFTWARETYPES.CommandLineApplication
            elif rawentrypoint.group == "gui_scripts":
                interfacetype = SOFTWARETYPES.DesktopApplication
            else:
                continue
            if rawentrypoint.value:
                module_name = rawentrypoint.value.strip().split(':')[0]
                add_entrypoint(g, res, rawentrypoint.name, module_name, interfacetype, args)
            else:
                add_entrypoint(g, res, rawentrypoint.name, "", interfacetype, args)
            found = True

        test_and_set_library(g, res, packagename, found, args)

    if prevdir: os.chdir(prevdir)

#pylint: disable=W0621
def add_dependency(g: Graph, res: Union[URIRef, BNode], value: str, args: AttribDict):
    for dependency in splitdependencies(value):
        if 'extra ==' in dependency and args.no_extras:
            continue
        dependency, depversion = parsedependency(dependency.strip())
        if dependency and not dependency.startswith(('<','>','=','^')):
            print(f"Found dependency {dependency} {depversion}",file=sys.stderr)
            depres = URIRef(generate_uri(dependency+depversion.replace(' ',''),args.baseuri,"dependency")) #version number is deliberately in ID here!
            g.add((depres, RDF.type, SDO.SoftwareApplication))
            g.add((depres, SDO.identifier, Literal(dependency)))
            g.add((depres, SDO.name, Literal(dependency)))
            if args.exactplatformversion:
                g.add((depres, SDO.runtimePlatform, Literal("Python " + str(sys.version_info.major) + "." + str(sys.version_info.minor) + "." + str(sys.version_info.micro))))
            else:
                g.add((depres, SDO.runtimePlatform, Literal("Python 3")))
            if depversion:
                g.add((depres, SDO.version, Literal(depversion)))
            g.add((res, CODEMETA.softwareRequirements, depres))

def add_entrypoint(g: Graph, res: Union[URIRef, BNode], name: str, module_name: str, interfacetype: Union[BNode,URIRef], args: AttribDict):
    """Translate an entrypoint to a targetProduct"""
    description = None
    if module_name:
        try:
            module = importlib.import_module(module_name)
            description = module.__doc__
        except:
            pass
    version = g.value(res,SDO.version) or "snapshot"
    targetproduct = URIRef(generate_uri(name, baseuri=args.baseuri,prefix=get_last_component(str(interfacetype)).lower()) + "/" + str(version))
    g.add((targetproduct, RDF.type, interfacetype))
    g.add((targetproduct, SDO.name, Literal(name)))
    if description:
        g.add((targetproduct, SDO.description, Literal(description)))
    g.add((targetproduct, SOFTWARETYPES.executableName, Literal(name)))
    if args.exactplatformversion:
        g.add((targetproduct, SDO.runtimePlatform, Literal("Python " + str(sys.version_info.major) + "." + str(sys.version_info.minor) + "." + str(sys.version_info.micro))))
    else:
        g.add((targetproduct, SDO.runtimePlatform, Literal("Python 3")))
    g.add((res, SDO.targetProduct, targetproduct))


def test_and_set_library(g: Graph, res: Union[URIRef, BNode], packagename: str, found_entrypoints: bool, args: AttribDict):
    """Is this resource a software library? Set the necessary targetProduct if so"""
    islibrary = False
    isweb = False
    for (_,_, cat) in g.triples((res, SDO.applicationCategory,None)):
        islibrary = islibrary or str(cat).lower().find("libraries") != -1
        isweb = isweb or str(cat).lower().find("internet") != -1 or str(cat).lower().find("web") != -1 or str(cat).lower().find("www") != -1

    if (not found_entrypoints and not isweb) or islibrary:
        targetproduct = URIRef(generate_uri(packagename, baseuri=args.baseuri,prefix="softwarelibrary"))
        g.add((targetproduct, RDF.type, SOFTWARETYPES.SoftwareLibrary))
        g.add((targetproduct, SDO.name, Literal(packagename)))
        g.add((targetproduct, SOFTWARETYPES.executableName, Literal(re.sub(r"[-_.]+", "-", packagename).lower()))) #see https://python.github.io/peps/pep-0503/
        if args.exactplatformversion:
            g.add((targetproduct, SDO.runtimePlatform, Literal("Python " + str(sys.version_info.major) + "." + str(sys.version_info.minor) + "." + str(sys.version_info.micro))))
        else:
            g.add((targetproduct, SDO.runtimePlatform, Literal("Python 3")))

def test_and_set_identifier(g: Graph, res: Union[URIRef, BNode], args: AttribDict):
    """ensure 'identifier' is always set"""
    name = g.value(res, SDO.name)
    if name and (res, SDO.identifier, None) not in g:
        g.set((res, SDO.identifier, Literal(str(name).lower())))



#!/usr/bin/env python3

import os
from setuptools import setup
try:
    from codemeta.codemeta import CodeMetaCommand
    cmdclass={
        'codemeta': CodeMetaCommand,
    }
except ImportError:
    cmdclass={}

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname),'r',encoding='utf-8').read()

setup(
    name = "CodeMetaPy",
    version = "2.4.1", #also adapt in codemeta.json
    author = "Maarten van Gompel",
    author_email = "proycon@anaproy.nl",
    description = ("Generate and manage CodeMeta software metadata"),
    license = "GPL-3.0-only",
    keywords = [ "software metadata", "codemeta", "schema.org", "rdf", "linked data"],
    url = "https://github.com/proycon/codemetapy",
    packages=['codemeta', 'codemeta.parsers','codemeta.serializers'],
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: Software Development",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: POSIX",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    zip_safe=False,
    include_package_data=True,
    package_data = { 'codemeta': ['schema/crosswalk.csv', 'schema/codemeta.jsonld', 'templates/*.html','resources/*.css', 'resources/fa-*' ] },
    install_requires=[ 'nameparser','importlib_metadata','BeautifulSoup4', 'rdflib >= 6.1.1','pyshacl', 'requests','lxml','pyyaml','Jinja2','pep517','tomlkit','pyproject_parser'],
    entry_points = {    'console_scripts': [ 'codemetapy = codemeta.codemeta:main' ] },
    cmdclass=cmdclass
)



import sys
import json
from codemeta.common import SDO, get_subgraph
from typing import Union, IO, Sequence
from rdflib import Graph, URIRef, BNode, Literal


def serialize_to_turtle(g: Graph, res: Union[Sequence,URIRef,None]) -> str:
    """Serializes the RDF graph to Turtle"""
    if res:
        #Get the subgraph that focusses on this specific resource (may be multiple) 
        #TODO: this may not work well with ordered lists yet!!
        if isinstance(res, (list,tuple)):
            g = get_subgraph(g, res)
        else:
            g = get_subgraph(g, [res])

    g.bind('sdo', SDO)
    return g.serialize(format='turtle', auto_compact=True)



import sys
import json
import os
from rdflib import Graph, URIRef, BNode, Literal
from typing import Union, IO, Optional
from codemeta.common import  PREFER_URIREF_PROPERTIES, AttribDict, REPOSTATUS, license_to_spdx, SDO, CODEMETA, SCHEMA_SOURCE, CODEMETA_SOURCE, SCHEMA_LOCAL_SOURCE, SCHEMA_SOURCE, CODEMETA_LOCAL_SOURCE, CODEMETA_SOURCE, STYPE_SOURCE, STYPE_LOCAL_SOURCE, IODATA_SOURCE, IODATA_LOCAL_SOURCE, init_context, SINGULAR_PROPERTIES, generate_uri, bind_graph, DEVIANT_CONTEXT


def rewrite_context(context: Union[list,str], args: AttribDict) -> list:
    """Rewrite remote contexts to their local counterparts"""
    local_contexts = [ x[0] for x in init_context(args) ]
    if isinstance(context, list):
        for i, v in enumerate(context):
            if isinstance(v, str):
                if v.startswith(("https://schema.org", "http://schema.org", "//schema.org")) or v == SCHEMA_SOURCE:
                    context[i] = SCHEMA_LOCAL_SOURCE
                elif v.startswith("https://doi.org/10.5063/schema") or v == CODEMETA_SOURCE:
                    context[i] = CODEMETA_LOCAL_SOURCE
                elif v.startswith(STYPE_SOURCE):
                    context[i] = STYPE_LOCAL_SOURCE
                elif v.startswith(IODATA_LOCAL_SOURCE):
                    context[i] = IODATA_LOCAL_SOURCE
                elif v.startswith(("file://","//")) and v not in local_contexts:
                    raise Exception(f"Refusing to load non-authorized local context: {v}")

        #remove some legacy contexts which we may encounter but would choke on if parsed
        try:
            context.remove("https://github.com/CLARIAH/tool-metadata")
        except ValueError:
            pass
    elif isinstance(context, str):
        context = rewrite_context([context], args)
    #ammend context
    if SCHEMA_LOCAL_SOURCE not in context:
        context.append(SCHEMA_LOCAL_SOURCE)
    if STYPE_LOCAL_SOURCE not in context:
        context.append(STYPE_LOCAL_SOURCE)
    if IODATA_LOCAL_SOURCE not in context:
        context.append(IODATA_LOCAL_SOURCE)

    for key,value in DEVIANT_CONTEXT.items():
        if {key:value} not in context:
            context.append({key:value})
    return context

def parse_jsonld(g: Graph, res: Union[BNode, URIRef,None], file_descriptor: IO, args: AttribDict) -> Union[str,None]:
    data = json.load(file_descriptor)
    return parse_jsonld_data(g,res, data, args)


def find_main_id(data: dict)  -> Union[str,None]:
    """Find the main URI in the JSON-LD resource, if there is only one, return None otherwise"""
    if '@graph' in data and len(data['@graph']) == 1:
        root = data['@graph'][0]
    else:
        root = data

    for k in ('@id','id'):
        if k in root:
            return root[k]

    return None


def inject_uri(data: dict, res: URIRef):
    if '@graph' in data and len(data['@graph']) == 1:
        data['@graph'][0]["@id"] = str(res)
        if 'id' in data['@graph'][0]: del data['@graph'][0]['id']
        print(f"    Injected (possibly temporary) URI {res}",file=sys.stderr)
    elif '@graph' in data and len(data['@graph']) == 0:
        print("    NOTE: Graph is empty!",file=sys.stderr)
    elif '@graph' not in data:
        data["@id"] = str(res)
        if 'id' in data: del data['id']
        print(f"    Injected (possibly temporary) URI {res}",file=sys.stderr)
    else:
        raise Exception("JSON-LD file does not describe a single resource (did you mean to use --graph instead?)")


def compute_hash(g: Graph, s: Union[URIRef,BNode], history: Optional[set] = None) -> int:
    """Computes a content hash for all contents in a resource"""
    values = []
    if not history: 
        history = set(s)
    else:
        history.add(s)
    if isinstance(s, URIRef):
        values.append(s)
    for s,p,o in g.triples((s,None,None)):
        values += [p,o]
        if isinstance(o, (URIRef,BNode)) and o not in history and (o,None,None) in g:
            #recursion step
            values.append( compute_hash(g, o, history) )
    return hash(tuple(values))
        


def skolemize(g: Graph, baseuri: Optional[str] = None):
    """In-place skolemization, turns blank nodes into uris"""
    #unlike Graph.skolemize, this one is in-place and edits the same graph rather than returning a copy
    #also, if blank nodes have identical content, they receive the same stub ID based on a hash of the content

    if baseuri:
        authority = baseuri
        if authority[-1] != "/": authority += "/"
        basepath = "stub/"
    else:
        authority = "file://" #for compatibility with rdflib
        basepath = "/stub/"

    hashes = {}
    for s,p,o in g.triples((None,None,None)):
        if isinstance(s, BNode) and s not in hashes:
            hashes[s] = compute_hash(g,s)

    for s,p,o in g.triples((None,None,None)):
        if isinstance(s, BNode):
            g.remove((s,p,o))
            #skolemize using hashes
            s = URIRef(authority + basepath + "H" + "%016x" % hashes[s])
            g.add((s,p,o))
        if isinstance(o, BNode):
            g.remove((s,p,o))
            if o in hashes:
                #skolemize using hashes
                o = URIRef(authority + basepath + "H" + "%016x" % hashes[o])
            else:
                o = o.skolemize(authority=authority, basepath=basepath)
            g.add((s,p,o))

def correct_wrong_uris(g:Graph, baseuri: Optional[str]):
    """Certain Literals should be URIRefs when possible, and some URIRefs are misinterpreted by rdflib and should be Literals."""
    for s,p,o in g:
        new_obj = o
        if str(o).startswith("//"):
            #we interpret this as a schemeless URL and will blatantly assume HTTPS (which is the most common source when fetching info, but this may be wrong)
            if isinstance(o, Literal):
                new_obj = Literal("https:" + o)
            elif isinstance(o, URIRef):
                new_obj = URIRef("https:" + o)
        if p in PREFER_URIREF_PROPERTIES:
            #turn Literals into URIRef for properties that prefer a URIRef
            if isinstance(o, Literal) and str(o).startswith("http"):
                new_obj =  URIRef(str(o))

            #these often get misinterpreted if they're not URIs, because rdflib prepends its baseuri
            cwd = os.getcwd()
            prefixes = [baseuri, cwd + "/", "file://" +cwd + "/", cwd, "file://" +cwd, "file://"]
            for prefix in prefixes:
                if prefix and str(o).startswith(prefix):
                    new_obj =  Literal(str(o)[len(prefix):])
                    break
        #commit the change
        if new_obj != o:
            g.remove((s,p,o))
            g.add((s,p,new_obj))


def parse_jsonld_data(g: Graph, res: Union[BNode, URIRef,None], data: dict, args: AttribDict, baseuri: Optional[str] = None) -> Union[str,None]:
    #preprocess json
    if '@context' not in data:
        data['@context'] = [ x[0] for x in init_context(args) ] + [DEVIANT_CONTEXT]
        print("    NOTE: Not a valid JSON-LD document, @context missing! Attempting to inject automatically...", file=sys.stderr)
    else:
        #rewrite context using the local schemas (also adds DEVIANT_CONTEXT)
        data['@context'] = rewrite_context(data['@context'], args)

    founduri = find_main_id(data)
    if founduri:
        print(f"    Found main resource with URI {founduri}",file=sys.stderr)
    if isinstance(res, URIRef) and founduri != str(res):
        #we're handling a single resource. Inject our own URI prior to parsing with rdflib
        inject_uri(data, res)

    #reserialize after edits
    reserialised_data: str = json.dumps(data, indent=4)

    #parse as RDF, add to main graph, and skolemize (turn blank nodes into URIs)
    skolemize(g.parse(data=reserialised_data, format="json-ld", publicID=baseuri if baseuri else args.baseuri), args.baseuri)
    correct_wrong_uris(g, args.baseuri)

    return founduri #return found uri (if any)
